
from six.moves.queue import Queue as ThreadQueue
from multiprocessing import Manager as ProcessManager
from nbdler.struct.signal import ID_TASK_FINISH, ID_TASK_STOP, ID_TASK_EXCEPTION, ID_TASK_PAUSE, \
    SIGNAL_TASK_PAUSE, Signal, ID_TASK_OPEN, SIGNAL_TASK_OPEN, SIGNAL_TASK_START, ID_TASK_START, \
    SIGNAL_TASK_FAIL, ID_TASK_FAIL, SIGNAL_CALLBACK_END, ID_CALLBACK_END
from nbdler.struct.misc import SignalQueue, ProcessInfo
from nbdler.utils.process_interface import make_class, make_method
from nbdler.utils.thread import ThreadCollector, Lock
from nbdler.struct.task import TaskWrapper, TaskQueue
from nbdler.handler import dlopen
from nbdler.exception import HandlerError, URLError, ClientError, \
    ChildProcessDisable


__all__ = ['Manager', 'manager', 'ManagerForChildProcess']


SIGNAL_CALLBACK_END = SIGNAL_CALLBACK_END(Signal(id=-1))
DEFAULT_MAX_TASK = 2


def manager(max_task=DEFAULT_MAX_TASK, open_args=(), open_kwargs=None, deamon=False, *, child_process=False):
    if child_process:
        manager_handler = ManagerForChildProcess
    else:
        manager_handler = Manager

    return manager_handler(max_task, dlopen, open_args, open_kwargs, deamon)


class RequestWrapper(TaskWrapper):
    __slots__ = ()


class HandlerWrapper(TaskWrapper):
    __slots__ = ()


class ManagerWrapper(TaskWrapper):
    __slots__ = ()


GROUP_OPEN = 'Manager.open'
GROUP_OPEN_CTRL = 'Manager.opening_controller'
GROUP_CTRL_T = 'Manager.controller(t)'
GROUP_CTRL_P = 'Manager.controller(p)'
GROUP_DO_NEXT = 'Manager.Task.do_next'


class Manager:
    """
        Attribute:
            open_handle: A open-handle for handler.

            max_task: Specify the upper limit of running task at the same time.

            open_args: opener default args.

            open_kwargs: opener default kwargs.

            daemon: Default make client thread to be daemon thread if True.

    """
    def __init__(self, max_task, open_handle, open_args=(), open_kwargs=None, daemon=False):
        self._full_queue = []
        self._queue = TaskQueue(unopened=[], opening=[], ready=[], queue=[], running=[],
                                paused=[], finished=[], failed=[])
        self._open_handle = open_handle
        self._max_task = max_task
        if not open_kwargs:
            open_kwargs = {}

        self._open_args = open_args
        self._open_kwargs = open_kwargs

        self._process_mgr = None
        self._thread_queue = ThreadQueue()
        self._process_queue = None

        self._thread_collector = ThreadCollector(daemon)

        self._queue_started = False

        self._queue_lock = Lock()

        self._process_collector = {}

        self._closed = False

    def putrequest(self, request, enqueue=True):
        """ Put a Request into Manager queue.
            Return task id corresponding to the request.
        """
        if not self._process_queue:
            raise ChildProcessDisable('cannot put a request running on '
                                      'child process mode without enabling.')
        if request in self._full_queue:
            raise ValueError('request is already in the task queue.')
        index = len(self._full_queue)
        wrapper = RequestWrapper(id=index, source=request, callback=ThreadQueue(),
                                 child_process=request.child_process)
        self._full_queue.append(wrapper)
        self._queue.unopened.append(index)
        if enqueue:
            self.enqueue(index)

        return index

    def enable_child_process_mode(self):
        if not self._process_mgr:
            self._process_mgr = ProcessManager()
            self._process_queue = self._process_mgr.Queue()
            self._process_collector[-1] = self._process_mgr._process

    def remove(self, task_id):
        """ Remove task by task id. """
        self._border_check(task_id)
        self._full_queue[task_id] = None

    def puthandler(self, handler):
        """ Put a Handler into Manager queue.
            Return task id corresponding to the handler.

            Warning: Not supported in child_process mode.
        """
        if handler in self._full_queue:
            raise ValueError('handler is already in the task queue.')
        index = len(self._full_queue)
        wrapper = HandlerWrapper(id=index, source=handler)
        self._full_queue.append(wrapper)
        handler.install_external_callback_queue(SignalQueue(
            id=index, queue=self._thread_queue))
        self._queue.ready.append(index)
        return index

    def get(self, task_id):
        """ Return HandlerWrapper of the task specified by id.

            Warning: this method does not supported by running in child process mode.
        """
        return self._full_queue[task_id]

    def process_info(self):
        infos = {}
        for k, v in self._process_collector.items():
            infos[k] = ProcessInfo(name=v.name, pid=v.pid, ident=v.ident)
        return infos

    def start_all(self):
        for i in self._queue.ready:
            self.start(i)
        for i in self._queue.paused:
            self.start(i)

    def get_all(self):
        for i in self._full_queue:
            if i is not None:
                yield i

    def enqueue(self, task_id):
        """ Make task into the plan queue.

            When queue is started, manager would run the tasks
            in the queue automatically until queue is empty.
        """
        self._border_check(task_id)
        with self._queue_lock:
            if task_id in self._queue.finished:
                raise RuntimeError('cannot queue a finished task.')
            self.__remove_queue_task(task_id)
            self._queue.queue.append(task_id)

    def getinstspeed(self, task_id=None):
        if task_id is not None:
            if not self._opened_info_checking(task_id):
                return 0
            return self._full_queue[task_id].source.getinstspeed()
        else:
            speed = 0
            for i in self._queue.running:
                speed += self._full_queue[i].source.getinstspeed()
            return speed

    def getavgspeed(self, task_id=None):
        if task_id is not None:
            if not self._opened_info_checking(task_id):
                return 0
            return self._full_queue[task_id].source.getavgspeed()
        else:
            speed = 0
            for i in self._queue.running:
                speed += self._full_queue[i].source.getavgspeed()
            return speed

    def getincbyte(self, task_id=None):
        if task_id is not None:
            if not self._opened_info_checking(task_id):
                return 0
            return self._full_queue[task_id].source.get_go_inc()
        else:
            incbyte = 0
            for i in self._queue.running:
                incbyte += self._full_queue[i].source.get_go_inc()
            return incbyte

    def get_remain_time(self, task_id=None):
        if task_id is not None:
            if not self._opened_info_checking(task_id):
                return float('inf')
            return self._full_queue[task_id].source.get_time_left()
        else:
            remain = 0
            for i in self._queue.running:
                remain += self._full_queue[i].source.get_time_left()
            return remain

    def get_remain_byte(self, task_id=None):
        if task_id is not None:
            if not self._opened_info_checking(task_id):
                return float('inf')
            return self._full_queue[task_id].source.get_byte_left()
        else:
            remain = 0
            for i in self._queue.running:
                remain += self._full_queue[i].source.get_byte_left()
            return remain

    def getfileinfo(self, task_id):
        if not self._opened_info_checking(task_id):
            return None
        return self._full_queue[task_id].source.getfileinfo()

    def geturlinfo(self, task_id):
        if not self._opened_info_checking(task_id):
            return None
        return self._full_queue[task_id].source.geturlinfo_all()

    def start_queue(self):
        """ Start running plan queue.

            The max number of running task is limited by self._max_task.
        """
        self._queue_started = True
        cur_task_len = len(self._queue.running)
        cur_task_len += len(self._queue.running)
        if cur_task_len < self._max_task:
            for i in range(self._max_task - cur_task_len):
                if self._queue.queue:
                    task_id = self._queue.queue[0]
                    self.start(task_id)
                else:
                    self.stop_queue()
                    break

    def stop_queue(self):
        """ Stop running plan queue.

            This method does not stop the running/opening task.
        """
        self._queue_started = False

    def start(self, task_id):
        self._border_check(task_id)

        if type(self._full_queue[task_id]) is RequestWrapper:
            self._full_queue[task_id].after(self.start, args=(task_id,), force=True)
            self.__remove_queue_task(task_id)
            self._queue.opening.append(task_id)
            self.open(task_id, block=False)
        else:
            self._full_queue[task_id].source.start()
            self._send_ctrl_signal(SIGNAL_TASK_START, task_id)

        self._check_controller_state()

    def resume(self, task_id):
        """ Resume the failed task.

            Only make failed task out of failed queue, cannot reopen/rerun the task.
        """
        if task_id not in self._queue.failed:
            raise RuntimeError('task <%d> is no need to resume.' % task_id)

        self._queue.failed.remove(task_id)
        if type(self._full_queue[task_id]) is RequestWrapper:
            self._queue.unopened.append(task_id)
        else:
            self._queue.ready.append(task_id)

    def pause(self, task_id):
        self._border_check(task_id)
        self._send_ctrl_signal(SIGNAL_TASK_PAUSE, task_id)

    stop = pause

    def pause_all(self, *, block=True):
        for i in self._queue.running:
            self.pause(i)
        for i in self._thread_collector.get_group(GROUP_OPEN_CTRL):
            task_id = i.owner
            self.pause(task_id)

        if block:
            for i in self._queue.running:
                self._full_queue[i].source.join()

    stop_all = pause_all

    def open(self, task_id, *, block=True):
        self._border_check(task_id)
        self._check_controller_state()

        if type(self._full_queue[task_id]) is not RequestWrapper:
            raise TypeError('opening task should be a type of request.')

        if block:
            return self._open(task_id)
        else:
            self._thread_collector.put(self._open, GROUP_OPEN, args=(task_id,),
                                       owner=task_id).start()

    def open_all(self, *, sequential=False, block=True):
        for i in self._queue.unopened:
            self.open(i, block=sequential)

        if block:
            self._thread_collector.wait(GROUP_OPEN)

    def is_all_finish(self):
        if not self.is_idle():
            return False
        if self._queue.paused or self._queue.failed or self._queue.unopened:
            return False
        return True

    def is_idle(self):
        """ Return True if all running and opening task are over. """
        if self._thread_collector.get_all():
            return False
        if self._queue.running:
            return False
        return True

    def get_queue(self):
        return tuple(self._queue.queue)

    def get_unopened(self):
        return tuple(self._queue.unopened)

    def get_finished(self):
        return tuple(self._queue.finished)

    def get_running(self):
        return tuple(self._queue.running)

    def get_failed(self):
        return tuple(self._queue.failed)

    def get_paused(self):
        return tuple(self._queue.paused)

    def is_finish(self, task_id):
        self._border_check(task_id)
        return self._full_queue[task_id].source.is_finished()

    def close(self):
        if not self.is_idle():
            raise RuntimeError('cannot close a running manager.')

        if self._closed:
            raise RuntimeError('manager has already closed.')

        if self._process_mgr:
            self._process_mgr.shutdown()
            self._process_mgr = None
            del self._process_collector[-1]
        for k, v in self._process_collector.items():
            if not v._close:
                v.terminate()
        self._closed = True

    def _opening_control_thread(self, task_id, handler):
        callback = self._full_queue[task_id].callback
        while True:
            signal = callback.get()
            if signal.id == ID_TASK_OPEN:
                self._full_queue[task_id] = signal.content
                self._send_ctrl_signal(SIGNAL_TASK_OPEN, task_id)
                callback.put_nowait(SIGNAL_CALLBACK_END)
            elif signal.id == ID_TASK_PAUSE:
                handler.pause(block=False)
                self._send_ctrl_signal(SIGNAL_TASK_PAUSE, task_id)
                callback.put_nowait(SIGNAL_CALLBACK_END)
            elif signal.id == ID_TASK_FAIL:
                self._send_ctrl_signal(SIGNAL_TASK_FAIL, task_id, signal.content)
                callback.put_nowait(SIGNAL_CALLBACK_END)
            elif signal.id == ID_CALLBACK_END:
                while not callback.empty():
                    callback.get_nowait()
                    callback.task_done()
                break
            callback.task_done()

        callback.task_done()

    def _check_controller_state(self):
        if not self._thread_collector.get_group(GROUP_CTRL_T):
            self._thread_collector.put(
                self._manager_control_thread,
                GROUP_CTRL_T, args=(self._thread_queue,),
                owner=self._thread_queue).start()
        if not self._thread_collector.get_group(GROUP_CTRL_P):
            self._thread_collector.put(
                self._manager_control_thread,
                GROUP_CTRL_P, args=(self._process_queue,),
                owner=self._process_queue).start()

    def _manager_control_thread(self, queue):
        while True:
            signal = queue.get()
            task_id = signal.content.id
            content = signal.content.content
            if signal.id == ID_TASK_FINISH:
                self.__finished_task_handler(task_id)
            elif signal.id == ID_TASK_OPEN:
                self.__opened_task_handler(task_id)
            elif signal.id in (ID_TASK_EXCEPTION, ID_TASK_FAIL):
                self.__failed_task_handler(task_id, content)
            elif signal.id == ID_TASK_STOP:
                self.__stopped_task_handler(task_id)
            elif signal.id == ID_TASK_PAUSE:
                self.__paused_task_handler(signal.content)
            elif signal.id == ID_TASK_START:
                self.__started_task_handler(task_id)
            elif signal.id == ID_CALLBACK_END:
                break

            self._do_next(task_id)
            queue.task_done()
            self._idle_checking()

        queue.task_done()

    def _border_check(self, task_id):
        if task_id >= len(self._full_queue) or self._full_queue[task_id] is None:
            raise IndexError('task id is invalid.')

    def _open(self, task_id):
        wrapper = None
        dlreq_wrapper = self._full_queue[task_id]
        handler = self._open_handle(child_process=dlreq_wrapper.child_process,
                                    **self._open_kwargs)
        if dlreq_wrapper.child_process:
            self._process_collector[task_id] = handler._process
        open_callback = dlreq_wrapper.callback
        self._thread_collector.put(self._opening_control_thread, GROUP_OPEN_CTRL,
                                   args=(task_id, handler), owner=task_id).start()
        try:
            handler.open(dlreq_wrapper.source)
        except (HandlerError, URLError, ClientError) as e:
            open_callback.put_nowait(
                SIGNAL_TASK_FAIL(Signal(id=task_id, content=e))
            )
        except Exception as e:
            open_callback.put_nowait(
                SIGNAL_TASK_FAIL(Signal(id=task_id, content=e))
            )
            raise
        else:

            if handler.is_opened():
                if dlreq_wrapper.child_process:
                    queue = self._process_queue
                else:
                    queue = self._thread_queue

                wrapper = HandlerWrapper(id=task_id, source=handler,
                                         child_process=dlreq_wrapper.child_process)
                handler.install_external_callback_queue(SignalQueue(
                    id=task_id, queue=queue))
                wrapper.move_from(dlreq_wrapper)
                open_callback.put_nowait(
                    SIGNAL_TASK_OPEN(content=wrapper))

            else:
                open_callback.put_nowait(SIGNAL_TASK_PAUSE(Signal(id=task_id)))

        return wrapper

    def _pause_opening(self, task_id):
        flag = False
        open_threads = self._thread_collector.get_group(GROUP_OPEN_CTRL)
        for i in open_threads:
            if i.owner == task_id:
                flag = True
        if flag:
            wrapper = self._full_queue[task_id]
            if type(wrapper) is RequestWrapper:
                wrapper.callback.put_nowait(SIGNAL_TASK_PAUSE())

    def _opened_info_checking(self, task_id):
        return type(self._full_queue[task_id]) is not RequestWrapper

    def _send_ctrl_signal(self, signal_wrapper, task_id, content=None):
        signal = signal_wrapper(Signal(id=task_id, content=content))
        if self._full_queue[task_id].child_process:
            self._process_queue.put_nowait(signal)
        else:
            self._thread_queue.put_nowait(signal)

    def _close_controller(self):
        if self._thread_collector.get_group(GROUP_CTRL_T):
            self._thread_queue.put_nowait(SIGNAL_CALLBACK_END)
        if self._thread_collector.get_group(GROUP_CTRL_P):
            self._process_queue.put_nowait(SIGNAL_CALLBACK_END)

    def _idle_checking(self):
        if not self._queue.opening and not self._queue.running and \
                (not self._queue_started or (self._queue_started and not self._queue.queue)):
            self._close_controller()

    def _do_next(self, task_id):
        next_task = self._full_queue[task_id].get_next()
        if next_task:
            self._thread_collector.put(next_task.handler, GROUP_DO_NEXT, args=next_task.args,
                                       kwargs=next_task.kwargs, owner=task_id).start()

    def __remove_queue_task(self, task_id):
        if task_id in self._queue.running:
            self._queue.running.remove(task_id)
        elif task_id in self._queue.queue:
            self._queue.queue.remove(task_id)
        elif task_id in self._queue.paused:
            self._queue.paused.remove(task_id)
        elif task_id in self._queue.ready:
            self._queue.ready.remove(task_id)
        elif task_id in self._queue.unopened:
            self._queue.unopened.remove(task_id)
        elif task_id in self._queue.failed:
            self._queue.failed.remove(task_id)
        elif task_id in self._queue.finished:
            self._queue.finished.remove(task_id)
        elif task_id in self._queue.opening:
            self._queue.opening.remove(task_id)

    def __paused_task_handler(self, signal):
        task_id = signal.id
        if task_id in self._queue.running:
            self._full_queue[task_id].source.pause(block=False)
        elif task_id in self._queue.unopened:
            self._pause_opening(task_id)

        if task_id in self._queue.running:
            self._queue.running.remove(task_id)
            self._queue.paused.append(task_id)

        if self._queue_started:
            self.start_queue()

    def __check_process_collector(self, task_id):
        if task_id in self._process_collector:
            del self._process_collector[task_id]

    def __opened_task_handler(self, task_id):
        self.__remove_queue_task(task_id)
        self._queue.ready.append(task_id)

    def __finished_task_handler(self, task_id):
        self.__remove_queue_task(task_id)
        self._queue.finished.append(task_id)

        self._full_queue[task_id].source.close()
        self.__check_process_collector(task_id)

        if self._queue_started:
            self.start_queue()

    def __failed_task_handler(self, task_id, exception):
        self.__remove_queue_task(task_id)
        self._full_queue[task_id].fail(exception)
        self._queue.failed.append(task_id)
        self._full_queue[task_id].reset()

        self.__check_process_collector(task_id)

        if self._queue_started:
            self.start_queue()

    def __stopped_task_handler(self, task_id):
        pass

    def __started_task_handler(self, task_id):
        self.__remove_queue_task(task_id)
        self._queue.running.append(task_id)


@make_class(Manager)
class ManagerForChildProcess:

    @make_method
    def putrequest(self, request, enqueue=True):
        pass

    # @make_method
    # def puthandler(self, handler):
    #     pass

    @make_method
    def remove(self, task_id):
        pass

    # @make_method
    # def get(self, task_id):
    #     pass

    # @make_method
    # def save_handle(self):
    #     pass

    @make_method
    def start_all(self):
        pass

    @make_method
    def get_all(self):
        pass

    @make_method
    def enqueue(self, task_id):
        pass

    @make_method
    def start_queue(self):
        pass

    @make_method
    def stop_queue(self):
        pass

    @make_method
    def start(self, task_id):
        pass

    @make_method
    def resume(self, task_id):
        pass

    @make_method
    def pause(self, task_id):
        pass

    stop = pause

    @make_method
    def pause_all(self, task_id):
        pass

    stop_all = pause_all

    @make_method
    def open(self, task_id, *, block=True):
        pass

    @make_method
    def open_all(self, *, sequential=False, block=True):
        pass

    @make_method
    def is_all_finish(self):
        pass

    @make_method
    def is_idle(self):
        pass

    @make_method
    def get_queue(self):
        pass

    @make_method
    def get_unopened(self):
        pass

    @make_method
    def get_finished(self):
        pass

    @make_method
    def get_running(self):
        pass

    @make_method
    def get_failed(self):
        pass

    @make_method
    def get_paused(self):
        pass

    @make_method
    def is_finish(self, task_id):
        pass

    @make_method
    def getfileinfo(self, task_id):
        pass

    @make_method
    def geturlinfo(self, task_id):
        pass

    @make_method
    def getincbyte(self, task_id=None):
        pass

    @make_method
    def getinstspeed(self, task_id=None):
        pass

    @make_method
    def getavgspeed(self, task_id=None):
        pass

    @make_method
    def get_remain_time(self, task_id=None):
        pass

    @make_method
    def get_remain_byte(self, task_id=None):
        pass

    @make_class
    def close(self):
        pass

    @make_method
    def enable_child_process_mode(self):
        pass

    @make_method
    def process_info(self):
        pass

    @make_method
    def dumps_handle(self):
        pass

    @make_method
    def loads_handle(self, dumped_data):
        pass






from collections import namedtuple
TaskQueue = namedtuple('_TaskQueue', 'unopened opening ready queue '
                                     'running paused finished failed')
Task = namedtuple('_Task', 'handler args kwargs')


class TaskWrapper:
    def __init__(self, id, source, callback=None, child_process=False):
        self.id = id
        self.source = source
        self.callback = callback
        self._after = []
        self._failed_exc = None
        self.child_process = child_process

    def after(self, handler, args=(), kwargs=None, force=False):
        if not kwargs:
            kwargs = {}
        task = Task(handler=handler, args=args, kwargs=kwargs)
        if force:
            self._after.insert(0, task)
        else:
            self._after.append(task)
        return self

    def get_next(self):
        if not self._after:
            return None
        return self._after.pop(0)

    def fail(self, msg):
        self._failed_exc = msg

    def move_from(self, task_wrapper):
        nexttask = task_wrapper.get_next()
        while nexttask:
            self._after.append(nexttask)
            nexttask = task_wrapper.get_next()

    def reset(self):
        self._after.clear()
        if self.callback:
            while not self.callback.empty():
                self.callback.get_nowait()








from nbdler.url.basic import BasicUrl, UrlInfo
from nbdler.url.response import UrlResponse
from wsgiref.headers import Headers
from nbdler.struct.misc import RangeFormat
from nbdler.struct.dump import SourceUrlDumpedData, UrlResponseDumpedData


class SourceUrl(BasicUrl):
    def __init__(self, id, url, headers=None, cookie=None,
                 proxy=None, max_conn=None, rangef=None, name=None):

        self.__origin_url = url
        self.id = id
        super(SourceUrl, self).__init__(url, headers)
        self.cookie = cookie
        self.proxy = proxy
        self.max_conn = max_conn

        if rangef is None:
            rangef = RangeFormat({'Range': 'bytes={begin}-{end_with}'})
        elif type(rangef) is dict:
            rangef = RangeFormat(rangef)
        elif type(rangef) is RangeFormat:
            pass
        else:
            raise TypeError()

        self._rangef = rangef
        self.name = name
        self._response = None

    def http_request_header(self, range):
        scheme, netloc, path, params, query, fragment = list(self._urlparse)

        query = query + ''.join([i + '=' + j for i, j in self._rangef.getquery(range).items()])
        query = query.lstrip('&')
        if query:
            path = path + '?' + query
        headers = Headers(self.headers.items())
        for i, j in self._rangef.getheader(range).items():
            headers.add_header(i, j)

        if self.cookie:
            headers.add_header('Cookie', self.cookie)

        path = path.strip().replace(' ', '%20')
        return path, headers

    def http_redirect(self, redurl):
        self.url = redurl
        return self

    def response(self, url, headers, code, length):
        self._response = UrlResponse(url=url, headers=dict(headers), code=code, length=length)

    def reset(self):
        self.url = self.__origin_url
        self._response = None

    def dump_data(self):
        return SourceUrlDumpedData(url=self.__origin_url, headers=dict(self.headers),
                                   cookie=self.cookie, proxy=self.proxy, max_conn=self.max_conn,
                                   rangef=dict(self._rangef), name=self.name,
                                   response=self._response and tuple(self._response.dump_data()))

    def load(self, dumped_data):
        data = UrlResponse(*dumped_data)
        self.response(data.url, data.headers, data.code, data.length)

    def getinfo(self):
        return UrlInfo(url=self.__origin_url, headers=list(self.headers.items()),
                       cookie=self.cookie, proxy=self.proxy, max_conn=self.max_conn,
                       rangef=dict(self._rangef), name=self.name, response=self._response)



from time import perf_counter


class AccumulatedTime:
    __slots__ = ('_start_time', '_inc_time', '_end_time')

    def __init__(self, inc_time=0):
        self._start_time = None
        self._inc_time = inc_time
        self._end_time = None

    def start(self):
        if not self._start_time:
            self._end_time = None
            self._start_time = perf_counter()

    def stop(self):
        if self._start_time:
            self._end_time = perf_counter()
            self._inc_time += self._end_time - self._start_time
            self._start_time = None

    def getinctime(self):
        return perf_counter() - self._start_time + self._inc_time if self._start_time else self._inc_time

    def clear(self):
        self._inc_time = 0


class InstSpeedMaker:
    __slots__ = ('_speed', '_prv_size', '_prv_time')

    def __init__(self):
        self._speed = 0
        self._prv_size = 0
        self._prv_time = None

    def start(self, start):
        self._prv_size = start
        self._prv_time = perf_counter()
        self._speed = 0

    def stop(self):
        self._prv_time = None
        self._speed = 0
        self._prv_size = 0

    def capture(self, cur_length):
        cur_time = perf_counter()
        self._speed = (cur_length - self._prv_size) / (cur_time - self._prv_time)
        self._prv_time = cur_time
        self._prv_size = cur_length

    def getspeed(self):
        return self._speed



SIMPLE_CHROMIUM_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                  'AppleWebKit/537.36 (KHTML, like Gecko) '
                  'Chrome/74.0.3729.169 Safari/537.36',
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate, br'
}

SIMPLE_FIREFOX_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:68.0) Gecko/20100101 Firefox/68.0',
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate, br'
}

__all__ = ['SIMPLE_CHROMIUM_HEADERS', 'SIMPLE_FIREFOX_HEADERS']


from collections import namedtuple

BlockDumpedData = namedtuple('_BlockDumpedData', 'rel_grid url_id progress')
ProgressDumpedData = namedtuple('_ProgressDumpedData', 'range go_inc done_inc')
HandlerDumpedData = namedtuple('_HandlerDumpedData', 'console url file maxthread buffsize blocksize')
FileDumpedData = namedtuple('_FileDumpedData', 'path name size')
UrlDumpedData = namedtuple('_UrlDumpedData', 'sources')
ConsoleDumpedData = namedtuple('_ConsoleDumpedData', 'block_data acum_time')
SourceUrlDumpedData = namedtuple('_SourceUrlDumpedData',
                                 'url headers cookie proxy max_conn rangef name response')
UrlResponseDumpedData = namedtuple('_UrlResponseDumpedData',
                                   'url headers code length')
ManagerDumpedData = namedtuple('_ManagerDumpedData', 'task_queue max_task')
TaskDumpedData = namedtuple('_TaskDumpedData', 'id filepath opened finished failed fileinfo request child_process')

RequestDumpedData = namedtuple('_RequestDumpedData',
                               'sources filepath resume block_size max_retries max_thread child_process')



from math import ceil
from collections import namedtuple
from nbdler.utils.thread import RLock
from nbdler.struct.dump import BlockDumpedData

try:
    xrange
except NameError:
    xrange = range

GridCell = namedtuple('GridCell', 'id begin end')


class Block:
    def __init__(self, content, progress, blocksize, abs_grid=None, rel_grid=None):
        self._content = content
        self._prog = progress
        self._blocksize = blocksize

        self._range = (int(progress.range[0] / blocksize),
                       int(ceil(progress.range[1] / blocksize)) if progress.range[1] else 1)
        if abs_grid:
            rel_grid = list([GridCell(id=i.id, begin=i.begin - self.begin, end=i.end - self.begin)
                             for i in abs_grid])
        if not rel_grid:
            rel_grid = [GridCell(id=None, begin=0, end=self.length)]

        self._rel_grid = rel_grid
        self._correct_lock = RLock()
        self.calibration()

    @property
    def begin(self):
        return self._range[0]

    @property
    def end(self):
        return self._range[1]

    @property
    def length(self):
        try:
            return self._range[1] - self._range[0]
        except TypeError:
            return float('inf')

    @property
    def grid(self):
        """ Grid's cells with relative position. """
        self.calibration()
        return self._rel_grid

    @property
    def abs_grid(self):
        """ Grid's cells with absolute position. """
        return [GridCell(id=c.id, begin=self.begin+c.begin, end=self.end+c.begin) for c in self.grid]

    @property
    def handler(self):
        return self._content.run

    @property
    def margin(self):
        margin_cell = self.grid[-1]
        try:
            if margin_cell.id is None:
                h = margin_cell.end - margin_cell.begin
            else:
                h = 0
        except TypeError:
            h = float('inf')
        return h

    def calibration(self):
        if self._blocksize != float('inf'):
            with self._correct_lock:
                self._correct_range()
                self._correct_grid()

    def __contains__(self, item):
        return item == self._content

    def _correct_grid(self):
        prog = self._content.getprogress()
        source = self._content.getsource()
        x_end = int(ceil(prog.go_inc / self._blocksize))
        margin_cell = self._rel_grid[-1]
        new_margin = GridCell(id=None, begin=x_end, end=self.length)
        if len(self._rel_grid) > 1:
            prv_cell = self._rel_grid[-2]
            if prv_cell.id == source.id:
                self._rel_grid[-2] = GridCell(id=source.id, begin=prv_cell.begin, end=x_end)
                if x_end == margin_cell.end:
                    self._rel_grid.pop(-1)
                else:
                    self._rel_grid[-1] = new_margin
            else:
                self._rel_grid.insert(-2, GridCell(id=source.id, begin=margin_cell.begin, end=x_end))
                self._rel_grid[-1] = new_margin
        else:
            self._rel_grid.insert(-2, GridCell(id=source.id, begin=0, end=x_end))
            self._rel_grid[-1] = new_margin

    def _correct_range(self):
        self._range = (self._range[0],
                       int(ceil(self._prog.end / self._blocksize)))

    def send_signal(self, signal):
        self._content.send_signal(signal)

    def release_buffer(self, fp):
        self._content.write_to_file(fp)

    clear_signal = property(lambda self: self._content.clear_signal, lambda self, v: None, None, '')

    getavgspeed = property(lambda self: self._prog.getavgspeed, lambda self, v: None, None, '')

    getinstspeed = property(lambda self: self._prog.getinstspeed, lambda self, v: None, None, '')

    getsource = property(lambda self: self._content.getsource, lambda self, v: None, None, '')

    clear = property(lambda self: self._prog.clear, lambda self, v: None, None, '')

    is_go_finished = lambda self: self._prog.go_finishflag

    is_finished = lambda self: self._prog.finishflag

    get_byte_left = lambda self: self._prog.go_remain

    get_go_inc = lambda self: self._prog.go_inc

    def __repr__(self):
        self.calibration()
        return "<%d>[%d-%d](%d-%d) - %s/%s" % (self.getsource().id, self.begin, self.end,
                                               self._prog.begin, self._prog.end, self.is_go_finished(),
                                               self.is_finished())

    def dump_data(self):
        return BlockDumpedData(rel_grid=tuple([tuple(i) for i in self.grid]), url_id=self.getsource().id,
                               progress=tuple(self._prog.dump_data()))






from nbdler.url.source import SourceUrl
from nbdler.exception import URLCrash, URLTimeout, URLUnknownError, NetworkDisconnected, MaxUsedExceedError
from nbdler.struct.dump import UrlDumpedData, SourceUrlDumpedData
from copy import copy

DEFAULT_CRASH_THRESHOLD = 10
DEFAULT_TIMEOUT_THRESHOLE = 50
DEFAULT_UNKNOWN_THRESHOLD = 150
DEFAULT_NETWORKDOWN_THRESHOLD = 10


class SourceWrapper:
    __slots__ = ('source', 'used', '_crash', '_timeout', '_unknown', '_network_down', '_max_used')

    def __init__(self, source, max_used):
        self.used = 0
        self._max_used = max_used
        self._crash = 0
        self._timeout = 0
        self._unknown = 0
        self._network_down = 0
        self.source = source

    def __contains__(self, item):
        return self.source.id == item.id

    def use_anyway(self):
        self.used += 1

    def use(self):
        self.used += 1
        if self._max_used is not None and self._max_used < self.used:
            self.used -= 1
            raise MaxUsedExceedError()

    def disuse(self):
        self.used -= 1

    def crash(self):
        self._crash += 1
        if DEFAULT_CRASH_THRESHOLD is not None and self._crash > DEFAULT_CRASH_THRESHOLD:
            raise URLCrash()

    def timeout(self):
        self._timeout += 1
        if DEFAULT_TIMEOUT_THRESHOLE is not None and self._timeout > DEFAULT_TIMEOUT_THRESHOLE:
            raise URLTimeout()

    def unknown(self):
        self._unknown += 1
        if DEFAULT_UNKNOWN_THRESHOLD is not None and self._unknown > DEFAULT_UNKNOWN_THRESHOLD:
            raise URLUnknownError()

    def network_down(self):
        self._network_down += 1
        if DEFAULT_NETWORKDOWN_THRESHOLD is not None and self._network_down > DEFAULT_NETWORKDOWN_THRESHOLD:
            raise NetworkDisconnected()

    def clear_counter(self):
        self._network_down = 0
        self._crash = 0
        self._timeout = 0
        self._unknown = 0

    def get(self):
        src = copy(self.source)
        src.reset()
        return src

    def is_available(self):
        try:
            if DEFAULT_CRASH_THRESHOLD is not None and self._crash > DEFAULT_CRASH_THRESHOLD:
                raise URLCrash()
            if DEFAULT_NETWORKDOWN_THRESHOLD is not None and self._network_down > DEFAULT_NETWORKDOWN_THRESHOLD:
                raise NetworkDisconnected()
            if DEFAULT_TIMEOUT_THRESHOLE is not None and self._timeout > DEFAULT_TIMEOUT_THRESHOLE:
                raise URLTimeout()
            if DEFAULT_UNKNOWN_THRESHOLD is not None and self._unknown > DEFAULT_UNKNOWN_THRESHOLD:
                raise URLUnknownError()
            if self._max_used is not None and self._max_used <= self.used:
                raise MaxUsedExceedError()
        except (NetworkDisconnected, URLCrash, URLTimeout, URLUnknownError, MaxUsedExceedError):
            return False

        return True

    def is_max_used(self):
        if self._max_used is not None and self._max_used <= self.used:
            return True
        return False


class Url:
    def __init__(self):
        self._sources = []

    def put(self, url, headers=None, cookie=None, *,
            proxy=None, max_conn=None, rangef=None, name=None):
        putid = self._newid()
        if name is None:
            name = str(putid)
        srcurl = SourceUrl(putid, url, headers, cookie, proxy, max_conn, rangef, name)
        self._sources[putid] = SourceWrapper(srcurl, max_conn)
        return srcurl

    def get(self, k, default=None):
        if type(k) is int:
            try:
                return self._sources[k]
            except ValueError:
                pass
        elif type(k) is str:
            for i in self._sources:
                if i.cont.name == k:
                    return i

        return default

    def get_all(self):
        return tuple(self._sources)

    def _newid(self):
        try:
            index = self._sources.index(None)
        except ValueError:
            index = len(self._sources)
            self._sources.append(None)
        return index

    def get_min_avl_used(self):
        min_max_used = sorted(self._sources, key=lambda i: i.used)
        for i in min_max_used:
            if i.is_available():
                return i
        return None

    def get_min_used(self):
        min_max_used = sorted(self._sources, key=lambda i: i.used)
        return min_max_used[0]

    def get_next(self, srcid):
        try:
            return self._sources[srcid + 1]
        except IndexError:
            return self._sources[0]

    def open_request(self, dlrequest):
        for v in dlrequest:
            self.put(url=v.url, headers=v.headers, cookie=v.cookie,
                     proxy=v.proxy, max_conn=v.max_conn,
                     rangef=v.rangef, name=v.name)

    def is_all_crashed(self):
        for i in self._sources:
            if i.is_available():
                return False

        return True

    def getwrapper(self, source, default=None):
        for i in self._sources:
            if source in i:
                return i
        return default

    def clear_counter(self):
        for i in self._sources:
            i.clear_counter()

    def dump_data(self):
        urls_data = []
        for wrapper in self._sources:
            urls_data.append(tuple(wrapper.source.dump_data()))
        return UrlDumpedData(sources=urls_data)

    def load(self, dumped_data):
        data = UrlDumpedData(*dumped_data)
        for source_data in data.sources:
            srcdata = SourceUrlDumpedData(*source_data)
            source = self.put(srcdata.url, srcdata.headers, srcdata.cookie,
                              proxy=srcdata.proxy, max_conn=srcdata.max_conn,
                              rangef=srcdata.rangef, name=srcdata.name)
            if srcdata.response:
                source.load(srcdata.response)




from nbdler.url.basic import UrlInfo
from nbdler.struct.dump import RequestDumpedData
from nbdler.collect import SIMPLE_CHROMIUM_HEADERS
DEFAULT_BLOCK_SIZE = 512 * 1024
DEFAULT_OPEN_MAX_RETRIES = 5


class Request:
    def __init__(self, url=None, headers=SIMPLE_CHROMIUM_HEADERS, cookie=None,
                 *, proxy=None, max_conn=None, range_format=None, name='main',
                 filepath='', block_size=DEFAULT_BLOCK_SIZE, max_retries=DEFAULT_OPEN_MAX_RETRIES,
                 max_thread=None, child_process=False, resume=False):

        self._sources = []
        self.filepath = filepath
        self.block_size = block_size
        self.max_retries = max_retries
        self.max_thread = max_thread
        self.child_process = child_process
        self.resume = resume
        if url:
            self.put(url, headers, cookie, proxy=proxy,
                     max_conn=max_conn, rangef=range_format, name=name)

    def put(self, url, headers=SIMPLE_CHROMIUM_HEADERS, cookie=None, *,
            proxy=None, max_conn=None, rangef=None, name=None):
        srcurl = UrlInfo(url=url, headers=headers, cookie=cookie, proxy=proxy,
                         max_conn=max_conn, rangef=rangef, name=str(name))
        self._sources.append(srcurl)
        return srcurl

    def __iter__(self):
        return iter(self._sources)

    def dump_data(self):
        sources = []
        for i in self:
            sources.append(tuple(i))
        return RequestDumpedData(sources=sources, resume=self.resume, block_size=self.block_size,
                                 max_retries=self.max_retries, max_thread=self.max_thread,
                                 child_process=self.child_process)

    @staticmethod
    def loads(dumped_data):
        data = RequestDumpedData(*dumped_data)

        req = Request(filepath=data.filepath, resume=data.resume, max_retries=data.max_retries,
                      child_process=data.child_process, max_thread=data.max_thread,
                      block_size=data.block_size)
        for i in data.sources:
            req.put(**dict(data))

        return req




from __future__ import division
import gc
from math import ceil
from time import sleep

from six.moves.queue import Queue
from nbdler.struct.block import Block, GridCell
from nbdler.utils.thread import ThreadCollector, Lock
from nbdler.struct.progress import Progress
from nbdler.client import build_client
from nbdler.console.abstract_console import AbstractConsole
from nbdler.struct.signal import SIGNAL_TASK_STOP, SIGNAL_TASK_SLICE, ID_TASK_BUFF, \
    ID_TASK_SLICE, ID_TASK_STOP, SIGNAL_TASK_FINISH, ID_THREAD_END, ID_TASK_FINISH, ID_TASK_PAUSE, SIGNAL_TASK_PAUSE, \
    ID_GAIERROR, ID_CRASH, ID_TIMEOUT, ID_UNKNOWN, ID_URL_STATUS, ID_NORMAL, SIGNAL_EXCEPTION, \
    SIGNAL_SWITCH, SIGNAL_WAIT, Signal, ID_CALLBACK_END, SIGNAL_CALLBACK_END

from nbdler.exception import URLUnknownError, URLTimeout, URLCrash, \
    NetworkDisconnected
from nbdler.struct.dump import BlockDumpedData, ProgressDumpedData, ConsoleDumpedData
from nbdler.struct.time_speed import AccumulatedTime, InstSpeedMaker

GROUP_CLIENT = 'Console.client'
GROUP_CONTROLLER = 'Console.controller'
GROUP_RELEASE = 'Console.release'
GROUP_SLICER = 'Console.slice'
GROUP_INSTSPEED = 'Console.inst_capture'

INST_SPEED_REFRESH_FREQUENCY = 3    # 3 Hz


class MultiClientConsole(AbstractConsole):
    def __init__(self, file, url, buffsize, block_size, maxthread, exception_callback):
        self._blocks = []
        self._block_size = block_size

        self._mxthread = maxthread
        self._buffsize = buffsize

        self.__total = int(ceil(file.getsize() / self._block_size))

        self.file = file
        self._url = url

        self._tpm = ThreadCollector()

        self._client_callback_queue = Queue()
        self.__exception_signal = exception_callback

        self.__buff_counter = 0

        self._buff_lock = Lock()
        self.runflag = False

        self._acum_time = AccumulatedTime(0)
        self._inst_maker = InstSpeedMaker()
        self._saver = None

        self._external_callback_queue = None

    def insert(self, content, progress, abs_grid=None, rel_grid=None):
        block = Block(content, progress, self._block_size, abs_grid=abs_grid, rel_grid=rel_grid)

        for i, v in enumerate(self._blocks):
            if v.begin > block.begin:
                self._blocks.insert(i, block)
                break
        else:
            self._blocks.append(block)

        return block

    def locate(self, block_index):
        c = 0
        for i in self._blocks:
            c += i.length
            if block_index < c:
                return i

    def getfullmap(self):
        for b in self._blocks:
            for i in b.abs_grid:
                yield i

    def __iter__(self):
        return self.getfullmap()

    def __getitem__(self, item):
        if type(item) is not int:
            raise ValueError()

        if item >= self.__total:
            raise IndexError()

        c = 0
        for i in self._blocks:
            c += i.length
            if item < c:
                return i.grid[item - c + i.length]

        raise IndexError()

    def prepare(self):

        self.file.open()

    def run(self):
        """ Run download task. """
        if not self._blocks:
            raise RuntimeError('cannot find any download block in the console.')
        self.runflag = True
        self._acum_time.start()
        self._inst_maker.start(self.get_go_inc())
        for block in self._blocks:
            block.clear_signal()
            self._tpm.put(block.handler, GROUP_CLIENT, args=(self._client_callback_queue,),
                          owner=block)
        self._tpm.start_group(GROUP_CLIENT)

        self._tpm.put(self._client_callback_handle_thread, GROUP_CONTROLLER).start()
        self._tpm.put(self._inst_speed_capture_thread, GROUP_INSTSPEED).start()

        if self._mxthread > self.get_client_cnt():
            self._make_slice_request()

    def get_block_size(self):
        """ Return console's unit block size. """
        return self._block_size

    def get_go_inc(self):
        """ Return byte size downloaded. """
        inc = 0
        for v in self._blocks:
            inc += v.get_go_inc()
        return inc

    def pause(self, block=True, timeout=None):
        """ Pause downloading. block until finish pausing only when block == True. """
        self.runflag = False
        self._client_callback_queue.put_nowait(SIGNAL_TASK_PAUSE())
        if block:
            self._tpm.join(timeout=timeout)
    stop = pause

    def getavgspeed(self):
        """ Return average download speed. """
        return self.get_go_inc() / (self._acum_time.getinctime() or float('inf'))

    def getinstspeed(self):
        """ Return instant download speed. """
        return self._inst_maker.getspeed()

    def _inst_speed_capture_thread(self):
        while self.runflag:
            self._inst_maker.capture(self.get_go_inc())
            freq = INST_SPEED_REFRESH_FREQUENCY
            if freq > 0:
                sleep(1 / freq)
            else:
                break

    def get_time_left(self):
        """ Return remaining time to finish the download task. """
        speed = self.getavgspeed()
        return self.get_byte_left() / speed if speed else float('inf')

    def get_byte_left(self):
        """ Return remaining byte to finish the download task. """
        return self.file.getsize() - self.get_go_inc()

    def get_online_cnt(self):
        """ Return the number of running client-threads. """
        return len(self._tpm.get_group(GROUP_CLIENT))

    def get_client_cnt(self):
        """ Return the number of running clients. """
        acum = 0
        for v in self._blocks:
            if not v.is_go_finished():
                acum += 1
        return acum

    def is_go_finished(self):
        """ Return True if download finished (finish writing to file is not necessary). """
        for v in self._blocks:
            if not v.is_go_finished():
                return False
        return self._check_complete()

    def is_finished(self):
        """ Return True if download task finished. """
        for v in self._blocks:
            if not v.is_finished():
                return False
        return self._check_complete()

    def _client_callback_handle_thread(self):
        """ A thread handler about callback signal. """
        while True:
            signal = self._client_callback_queue.get()
            if signal.id == ID_TASK_BUFF:
                self.__buffer_signal_handler(signal.content)
            elif signal.id == ID_TASK_SLICE:
                self.__slice_feedback_handler(signal.content)
            elif signal.id == ID_THREAD_END:
                self.__client_thread_end_handler(signal.content)
            elif signal.id == ID_URL_STATUS:
                self.__url_status_handler(signal.content)
            elif signal.id == ID_TASK_PAUSE:
                self.__pause_handler()
            elif signal.id == ID_TASK_FINISH:
                self.__finish_handler()
            elif signal.id == ID_TASK_STOP:
                self.__stop_handler()
            elif signal.id == ID_CALLBACK_END:
                break

            self._client_callback_queue.task_done()

        self._client_callback_queue.task_done()

    def _buffer_release_thread(self):
        """ A thread handling writing buffer to file. """
        with self._buff_lock:
            for v in self._blocks:
                v.release_buffer(self.file)
            if self._saver:
                self._saver.dump()
        gc.collect()

    def _find_block_from_client(self, client, defalut=None):
        """ Find the block"""
        for v in self._blocks:
            if client in v:
                return v
        return defalut

    def _make_slice_request(self):
        """ Make a slice request. """
        maxblock = sorted(self._blocks, key=lambda i: i.margin, reverse=True)[0]
        margin_len = maxblock.margin
        put_begin = maxblock.begin + (maxblock.length - margin_len) + int(ceil(margin_len / 2))
        put_end = maxblock.end
        signal = SIGNAL_TASK_SLICE((put_begin * self._block_size, put_end * self._block_size))
        maxblock.send_signal(signal)

    def _check_complete(self):
        """ Return True if total blocks length completely match. """
        acum_len = 0
        for v in self._blocks:
            v.calibration()
            acum_len += v.length
        return self.__total == acum_len

    def __client_thread_end_handler(self, client):
        """ Callback handler when signal.id == ID_THREAD_END:
                <CLIENT THREAD END OF LIFE>
        """
        block = self._find_block_from_client(client)
        srcwrapper = self._url.getwrapper(block.getsource())
        if self.runflag:
            if block.is_go_finished():
                if self._mxthread > self.get_client_cnt():
                    self._make_slice_request()
                srcwrapper.disuse()

            else:
                self._tpm.put(client.run, GROUP_CLIENT, args=(self._client_callback_queue,),
                              owner=block).start()

        if self.is_go_finished():
            self._buffer_release_thread()
            self._finish_task()

    def _finish_task(self):
        self._client_callback_queue.put_nowait(SIGNAL_TASK_FINISH())

        if self._external_callback_queue:
            self._external_callback_queue.queue.put_nowait(
                SIGNAL_TASK_FINISH(
                    Signal(id=self._external_callback_queue.id)
                )
            )

    def _failed_task(self, exception):
        self.__exception_signal.put_nowait(SIGNAL_EXCEPTION(exception))

        if self._external_callback_queue:
            self._external_callback_queue.queue.put_nowait(
                SIGNAL_EXCEPTION(
                    Signal(id=self._external_callback_queue.id,
                           content=exception)
                )
            )

    def __stop_handler(self):
        self._buffer_release_thread()
        self.file.close()
        self._acum_time.stop()
        self._inst_maker.stop()
        self.runflag = False
        self._client_callback_queue.put_nowait(SIGNAL_CALLBACK_END())
        self.__exception_signal.put_nowait(SIGNAL_TASK_STOP())
        if self._external_callback_queue:
            self._external_callback_queue.queue.put_nowait(
                SIGNAL_TASK_STOP(Signal(id=self._external_callback_queue.id))
            )

    def __finish_handler(self):
        self._client_callback_queue.put_nowait(SIGNAL_TASK_STOP())

    def __slice_feedback_handler(self, slice):
        """ Callback handler when signal.id == ID_SLICE:
                <SLICE FEEDBACK>
        """
        if slice.range:
            srcwrapper = self._url.get_min_avl_used()
            if not srcwrapper:
                srcwrapper = self._url.get_min_used()

            prog = Progress(slice.range)
            client = slice.client(srcwrapper.get(), prog)
            block = self.insert(client, prog)
            srcwrapper.use()
            if self.runflag:
                self._tpm.put(block.handler, GROUP_CLIENT,
                              args=(self._client_callback_queue,),
                              owner=block).start()

        if self._mxthread > self.get_client_cnt():
            self._make_slice_request()

    def __buffer_signal_handler(self, byte):
        """ Callback handler when signal.id == ID_BUFF:
                <BUFFER COUNTER>
        """
        self._url.clear_counter()
        self.__buff_counter += byte
        if self.__buff_counter >= self._buffsize:
            self.__buff_counter = 0
            self._tpm.put(self._buffer_release_thread, GROUP_RELEASE).start()

    def __url_status_handler(self, url_status):
        """ Callback handler when signal.id == ID_URL_STATUS:
                <URL SIGNAL>
        """
        client = url_status.content.client
        exception = url_status.content.exception
        source = client.getsource()
        srcwrapper = self._url.getwrapper(source)
        block = self._find_block_from_client(client)
        try:
            if url_status.id == ID_NORMAL:
                srcwrapper.clear_counter()
            elif url_status.id == ID_TIMEOUT:
                srcwrapper.timeout()
            elif url_status.id == ID_CRASH:
                srcwrapper.crash()
            elif url_status.id == ID_UNKNOWN:
                srcwrapper.unknown()
            elif url_status.id == ID_GAIERROR:
                srcwrapper.network_down()
        except (URLCrash, URLTimeout, URLUnknownError, NetworkDisconnected) as e:

            srcwrapper.disuse()
            nextwrapper = self._url.get_min_avl_used()
            if not nextwrapper:
                nextwrapper = self._url.get_next(srcwrapper.source.id)

                if self._url.is_all_crashed():
                    self._failed_task(e)
                else:
                    block.send_signal(SIGNAL_WAIT(1))

            nextwrapper.use_anyway()

            block.send_signal(SIGNAL_SWITCH(nextwrapper.get()))

    def __pause_handler(self):
        """ Callback handler when signal.id == ID_PAUSE:
                <PAUSE SIGNAL>
        """
        for thread in self._tpm.get_group(GROUP_CLIENT):
            thread.owner.send_signal(SIGNAL_TASK_PAUSE())

        self._tpm.wait(GROUP_SLICER)
        self._tpm.wait(GROUP_CLIENT)
        self._tpm.wait(GROUP_RELEASE)

        self._client_callback_queue.put_nowait(SIGNAL_TASK_STOP())

        if self._external_callback_queue:
            self._external_callback_queue.queue.put_nowait(
                SIGNAL_TASK_PAUSE(
                    Signal(id=self._external_callback_queue.id)
                )
            )

    def dump_data(self):
        return ConsoleDumpedData(block_data=list([tuple(b.dump_data()) for b in self._blocks]),
                                 acum_time=self._acum_time.getinctime())

    def load(self, dumped_data):
        data = ConsoleDumpedData(*dumped_data)
        self._acum_time = AccumulatedTime(data.acum_time)

        for i in data.block_data:
            b = BlockDumpedData(*i)
            p = ProgressDumpedData(*b.progress)
            prog = Progress(p.range, p.go_inc, p.done_inc)
            srcwrapper = self._url.get(b.url_id)
            source = srcwrapper.get()
            client = build_client(source, prog)
            srcwrapper.use()
            self.insert(client, prog,
                        rel_grid=list([GridCell(*i) for i in b.rel_grid]))

    def install_saver(self, saver):
        self._saver = saver

    def install_external_callback_queue(self, signal_queue):
        self._external_callback_queue = signal_queue

    def join(self, timeout=None):
        self._tpm.join(timeout)

    def getBuffCnter(self):
        return self.__buff_counter




from collections import namedtuple

Slice = namedtuple('_Slice', 'client range')
InitialResult = namedtuple('_InitialResult', 'client filename filesize unspecified_size response')
ClientException = namedtuple('_ClientException', 'client exception')
SignalQueue = namedtuple('SignalQueue', 'id queue')
ProcessInfo = namedtuple('ProcessInfo', 'name pid ident')


class RangeFormat:
    def __init__(self, rangef):
        self._full_formats = rangef

        self._queryf = {}
        self._headerf = {}

        for i, j in self._full_formats.items():
            if i[0] != '&':
                self._headerf[i] = j
            else:
                self._queryf[i] = j

    def _format(self, range, dict):
        retitem = dict.copy()
        for i, j in dict.items():
            begin = range[0]
            end = range[1] or ''
            end_with = range[1] - 1 if range[1] is not None and range[1] > 0 else ''
            length = (range[1] or 0) - range[0]
            retitem[i] = j.format(begin=begin, end=end, end_with=end_with, length=length)
        return retitem

    getheader = lambda self, range: RangeFormat._format(self, range, self._headerf)

    getquery = lambda self, range: RangeFormat._format(self, range, self._queryf)

    def __iter__(self):
        return iter(self._full_formats.items())




from __future__ import division
from nbdler.struct.dump import ProgressDumpedData
from nbdler.struct.time_speed import AccumulatedTime


class Progress:
    def __init__(self, _range, go_inc: int = 0, done_inc: int = 0):
        if len(_range) == 1:
            _range = (_range[0], None)
        self.range = _range

        if done_inc != go_inc:
            done_inc = go_inc
        self.go_inc = go_inc
        self.done_inc = done_inc

        self._acum_time = AccumulatedTime(0)

        self._retrieve_buff = []

    @property
    def go_finishflag(self):
        return self.go_inc >= self.length

    @property
    def done_finishflag(self):
        return self.done_inc >= self.length

    @property
    def finishflag(self):
        return self.go_finishflag and self.done_finishflag

    @property
    def begin(self):
        return self.range[0]

    @property
    def end(self):
        return self.range[1]

    @property
    def length(self):
        try:
            return self.range[1] - self.range[0]
        except TypeError:
            return float('inf')

    @property
    def go_remain(self):
        return self.length - self.go_inc

    @property
    def done_remain(self):
        return self.length - self.done_inc

    @property
    def differ(self):
        return self.go_inc - self.done_inc

    def getavgspeed(self):
        return self.go_inc / self._acum_time.getinctime()

    def go(self, byte_len):
        self.go_inc += byte_len

    def done(self, byte_len):
        self.done_inc += byte_len

    def buffer(self, b):
        self._retrieve_buff.append(b)

    def fetch_buffer(self):
        while True:
            try:
                buff = self._retrieve_buff.pop(0)
                self.done(len(buff))
                yield buff
            except IndexError:
                break

    def is_empty_buff(self):
        return self._retrieve_buff == []

    def close(self):
        self._acum_time.stop()

    def start(self):
        self._acum_time.start()

    def stop(self):
        self._acum_time.stop()

    def force_to_finish_go(self):
        self.range = (self.begin, self.begin + self.go_inc)

    def clear(self):
        self.go_inc = 0
        self.done_inc = 0
        self._retrieve_buff = []

    def slice_check(self, req_range):
        if req_range[0] > self.begin + self.go_inc:
            if req_range[1] != self.end:
                req_range = (req_range[0], self.end)
            if req_range[0] >= req_range[1]:
                req_range = ()
            put_range = req_range
        else:
            put_range = ()

        if put_range:
            self.range = (self.range[0], put_range[0])

        return put_range

    def __repr__(self):
        return "[%d-%d] %d/%d,finish=%s" % (self.begin, self.end, self.go_inc, self.length, self.finishflag)

    def dump_data(self):
        return ProgressDumpedData(range=self.range, go_inc=self.go_inc,
                                  done_inc=self.done_inc)







import gc
from time import sleep

from nbdler.utils.thread import ThreadCollector, Lock
from nbdler.struct.block import Block
from six.moves.queue import Queue
from nbdler.struct.dump import BlockDumpedData, ConsoleDumpedData
from nbdler.struct.time_speed import AccumulatedTime, InstSpeedMaker
from nbdler.struct.progress import Progress
from nbdler.client import build_client
from nbdler.console.abstract_console import AbstractConsole
from nbdler.struct.signal import Signal, SIGNAL_TASK_STOP, ID_TASK_BUFF, SIGNAL_CALLBACK_END, \
    ID_TASK_STOP, SIGNAL_TASK_FINISH, ID_THREAD_END, ID_TASK_FINISH, ID_TASK_PAUSE, SIGNAL_TASK_PAUSE, \
    ID_GAIERROR, ID_CRASH, ID_TIMEOUT, ID_UNKNOWN, ID_URL_STATUS, ID_NORMAL, SIGNAL_EXCEPTION

from nbdler.exception import URLUnknownError, URLTimeout, URLCrash, NetworkDisconnected


GROUP_CONTROLLER = 'Console.controller'
GROUP_CLIENT = 'Console.client'
GROUP_RELEASE = 'Console.release'
GROUP_SLICER = 'Console.slice'
GROUP_INSTSPEED = 'Console.inst_capture'


INST_SPEED_REFRESH_FREQUENCY = 2    # 2 Hz


class SingleClientConsole(AbstractConsole):
    def __init__(self, file, url, buffsize, block_size, maxthread, exception_callback):
        self._block_size = block_size

        self._buffsize = buffsize
        self._block = None
        self.file = file
        self._url = url
        self._tpm = ThreadCollector()
        self._client_callback_queue = Queue()
        self.__exception_signal = exception_callback

        self.__buff_counter = 0
        self._buff_lock = Lock()
        self.runflag = False

        self._acum_time = AccumulatedTime(0)
        self._inst_maker = InstSpeedMaker()
        self._saver = None

        self.__callback_queue = None

    def insert(self, client, progress):
        if self._block:
            raise RuntimeError('cannot insert more than one block into single client console.')
        self._block = Block(client, progress, self._block_size)

    def prepare(self):
        self.file.open()

    def run(self):
        self.runflag = True
        self._block.clear()
        self._acum_time.start()
        self._inst_maker.start(self.get_go_inc())
        self._tpm.put(self._block.handler, GROUP_CLIENT,
                      args=(self._client_callback_queue,), owner=self._block).start()
        self._tpm.put(self._client_callback_handle_thread, GROUP_CONTROLLER).start()
        self._tpm.put(self._inst_speed_capture_thread, GROUP_INSTSPEED).start()

    def __stop_handler(self):
        self.file.close()
        self._acum_time.stop()
        self._inst_maker.stop()
        self.runflag = False
        self._client_callback_queue.put_nowait(SIGNAL_CALLBACK_END())
        self.__exception_signal.put_nowait(SIGNAL_TASK_STOP())
        if self.__callback_queue:
            self.__callback_queue.queue.put_nowait(
                SIGNAL_TASK_STOP(Signal(id=self.__callback_queue.id))
            )

    def pause(self, block=True, timeout=None):
        self.runflag = False
        self._client_callback_queue.put_nowait(SIGNAL_TASK_PAUSE())
        if block:
            self._tpm.wait(GROUP_CONTROLLER, timeout=timeout)

    def is_finish_go(self):
        return self._block.is_go_finished()

    def get_byte_left(self):
        if self.is_finish_go():
            return 0
        else:
            return float('inf')

    def get_time_left(self):
        if self.is_finish_go():
            return 0
        else:
            return float('inf')

    def getavgspeed(self):
        return self.get_go_inc() / (self._acum_time.getinctime() or float('inf'))

    def get_go_inc(self):
        return self._block.get_go_inc()

    def getinstspeed(self):
        return self._inst_maker.getspeed()

    def _inst_speed_capture_thread(self):
        while self.runflag:
            self._inst_maker.capture(self.get_go_inc())
            freq = INST_SPEED_REFRESH_FREQUENCY
            if freq > 0:
                sleep(1 / freq)
            else:
                break

    def _client_callback_handle_thread(self):
        while True:
            signal = self._client_callback_queue.get()
            if signal.id == ID_TASK_BUFF:
                self.__buffer_signal_handler(signal.content)
            elif signal.id == ID_THREAD_END:
                self.__client_thread_end_handler(signal.content)
            elif signal.id == ID_URL_STATUS:
                self.__url_status_handler(signal.content)
            elif signal.id == ID_TASK_PAUSE:
                self.__pause_handler()
            elif signal.id == ID_TASK_FINISH:
                self.__stop_handler()
            elif signal.id == ID_TASK_STOP:
                break

            self._client_callback_queue.task_done()

        self._client_callback_queue.task_done()

    def is_finished(self):
        return self._block.is_finished()

    def _finish_task(self):
        self._client_callback_queue.put_nowait(SIGNAL_TASK_FINISH())

        if self.__callback_queue:
            self.__callback_queue.queue.put_nowait(
                SIGNAL_TASK_FINISH(
                    Signal(id=self.__callback_queue.id)
                )
            )

    def __pause_handler(self):
        """ Callback handler when signal.id == ID_PAUSE:
                <PAUSE SIGNAL>
        """
        for thread in self._tpm.get_group(GROUP_CLIENT):
            thread.owner.send_signal(SIGNAL_TASK_PAUSE())

        self._tpm.wait(GROUP_CLIENT)
        self._tpm.wait(GROUP_RELEASE)

        self._buffer_release_thread()
        self._client_callback_queue.put_nowait(SIGNAL_TASK_STOP())

        if self.__callback_queue:
            self.__callback_queue.queue.put_nowait(
                SIGNAL_TASK_PAUSE(
                    Signal(id=self.__callback_queue.id)
                )
            )

    def __url_status_handler(self, url_status):
        """ Callback handler when signal.id == ID_URL_STATUS:
                <URL SIGNAL>
        """
        client = url_status.content.client
        exception = url_status.content.exception
        source = client.getsource()
        srcwrapper = self._url.getwrapper(source)
        try:
            if url_status.id == ID_NORMAL:
                srcwrapper.clear_counter()
            elif url_status.id == ID_TIMEOUT:
                srcwrapper.timeout()
            elif url_status.id == ID_CRASH:
                srcwrapper.crash()
            elif url_status.id == ID_UNKNOWN:
                srcwrapper.unknown()
            elif url_status.id == ID_GAIERROR:
                srcwrapper.network_down()
        except (URLCrash, URLTimeout, URLUnknownError, NetworkDisconnected) as e:
            self.__exception_signal.put_nowait(SIGNAL_EXCEPTION(e))

    def _buffer_release_thread(self):
        """ A thread handler about releasing buffer. """
        with self._buff_lock:
            self._block.release_buffer(self.file)
            if self._saver:
                self._saver.dump()
        gc.collect()

    def __buffer_signal_handler(self, byte):
        """ Callback handler when signal.id == ID_BUFF:
                <BUFFER COUNTER>
        """
        self._url.clear_counter()
        self.__buff_counter += byte
        if self.__buff_counter >= self._buffsize:
            self.__buff_counter = 0
            self._tpm.put(self._buffer_release_thread, GROUP_RELEASE).start()

    def __client_thread_end_handler(self, client):
        """ Callback handler when signal.id == ID_THREAD_END:
                <CLIENT THREAD END OF LIFE>
        """
        if self.runflag:
            if self._block.is_go_finished():
                srcwrapper = self._url.getwrapper(self._block.getsource())
                srcwrapper.disuse()
            else:
                self._block.clear()
                self._tpm.put(client.run, GROUP_CLIENT, args=(self._client_callback_queue,),
                              owner=self._block).start()

        if self.is_finish_go():
            self._buffer_release_thread()
            self._finish_task()

    def install_saver(self, saver):
        self._saver = saver

    def dump_data(self):
        return ConsoleDumpedData(block_data=tuple(self._block.dump_data()),
                                 acum_time=self._acum_time.getinctime())

    def load(self, dumped_data):
        data = ConsoleDumpedData(*dumped_data)
        self._acum_time = AccumulatedTime(data.acum_time)

        for i in data.block_data:
            b = BlockDumpedData(*i)
            prog = Progress((0,))
            srcwrapper = self._url.get(b.url_id)
            source = srcwrapper.get()
            client = build_client(source, prog)
            srcwrapper.use()
            self.insert(client, prog)

    def get_block_size(self):
        return self._block_size

    def join(self, timeout=None):
        self._tpm.join(timeout)

    def install_external_callback_queue(self, handle):
        self.__callback_queue = handle

    def getBuffCnter(self):
        return self.__buff_counter




# ---------------------------------------------------------------------------
# Name:        nbdler/client/__init__.py
# Author:      ZSAIm
#
# Created:     01-Sept-2019
# License:     Apache-2.0
# ---------------------------------------------------------------------------

from nbdler.client import http

__all__ = ('build_initial_opener', 'build_client')


def build_initial_opener(source, progress, callback):
    if source.scheme in ('http', 'https'):
        handler = http.build_initial_opener
    else:
        raise ValueError('Unsupported protocol (%s).' % source.scheme)

    return handler(source, progress, callback)


def build_client(source, progress):
    if source.scheme in ('http', 'https'):
        handler = http.HTTPClient
    else:
        raise ValueError('Unsupported protocol (%s).' % source.scheme)

    return handler(source, progress)
